#!/usr/bin/python
from pathlib import Path
import argparse
import sys

from .__init__ import __version__

parser = argparse.ArgumentParser(prog="txrm2tiff", description="Converter of txrm/xrm files to OME tif/tiff files", add_help=False)
convert_group = parser.add_argument_group("Converter arguments")
convert_group.add_argument('-i', '--input', dest='input_path', default="", type=str, action='store', help='txrm or xrm file to convert, or directory containing txrm and/or xrm files')
convert_group.add_argument('-r', '--reference', dest='custom_reference', default="", type=str, action='store', help='specify a custom file (tiff, txrm or xrm) for reference (ignores internal reference)')
convert_group.add_argument('-o', '--output-path', dest='output_path', default=None, action='store', help='specify output (can be directory or file) - specified directories that do not exist will be created')
convert_group.add_argument('--ignore-ref', dest='ignore_reference', default=False, action='store_true', help='if specified ignore any internal reference')
convert_group.add_argument('--set-logging', dest='logging_level', default="info", type=str, action='store', help='pick logging level (options: debug, info, warning, error, critical)')

package_group = parser.add_argument_group("Package info")
package_group.add_argument('-v', '--version', action='version', version="%(prog)s {}".format(__version__))
package_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

setup_subparsers = parser.add_subparsers(title="Setup options", description="enter `txrm2tiff setup -h` for details")
setup_parser = setup_subparsers.add_parser(name="setup", add_help=False)
setup_parser.add_argument("-w", "--windows-shortcut", action='store_true', help="creates a Windows shortcut on the user's desktop that accepts drag and dropped files (allows batch processing)")
setup_info_group = setup_parser.add_argument_group("Setup info")
setup_info_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

def main():
    args = parser.parse_args()
    if hasattr(args, 'windows_shortcut'):
        if args.windows_shortcut:
            from .shortcut_creator import create_Windows_shortcut
            create_Windows_shortcut()
        else:
            setup_parser.print_help()
        return
    if len(sys.argv) > 1 and args.input_path:
        from .run import run
        if not args.custom_reference:
            args.custom_reference = None
        run(**vars(args))

    else:
        parser.print_help()



# -*- coding: utf-8 -*-

from xml.etree import ElementTree
import datetime
import logging
from pathlib import Path

from olefile import OleFileIO, isOleFile
import numpy as np
import tifffile as tf
from oxdls import OMEXML

from . import txrm_wrapper

DEFAULT_STR_DTYPE="uint16"
DEFAULT_NP_DTYPE = np.uint16


def _apply_reference(images, reference, np_dtype):
    floated_and_referenced = [(image * 100.) / reference for image in images]
    referenced_image = []
    for image in floated_and_referenced:
        if np.isnan(image).any() or np.isinf(image).any():
            logging.warning("Potential dead pixels found. "
                            "NaN was output for at least one pixel in the referenced image.")
        referenced_image.append(np.around(image).astype(np_dtype))
    return referenced_image


def _get_reference(ole, txrm_name, custom_reference, ignore_reference):
    if custom_reference is not None:
        logging.debug("%s is being referenced using %s and processed.", txrm_name, custom_reference.name)
        reference_path = str(custom_reference)
        try:
            if isOleFile(reference_path):
                with OleFileIO(reference_path) as ref_ole:
                    references = txrm_wrapper.extract_all_images(ref_ole)  # should be float for averaging & dividing
            elif ".tif" in reference_path:
                with tf.TiffFile(reference_path) as tif:
                    references = [page for page in tif.pages[:]]
            else:
                logging.error("Unable to open file '%s'. Only tif/tiff or xrm/txrm files are supported for custom references.", reference_path)
                raise IOError(f"Unable to open file '{reference_path}'. Only tif/tiff or xrm/txrm files are supported for custom references.")
        except:
            logging.error("Error occurred reading custom reference", exc_info=True)
            raise
        if len(references) > 1:
            # if reference file is an image stack take median of the images
            return np.median(np.asarray(references), axis=0)
        return references[0]

    elif ole.exists("ReferenceData/Image") and not ignore_reference:
        logging.debug("%s is being referenced and processed using an internal reference.", txrm_name)
        return txrm_wrapper.extract_reference_image(ole)

    logging.debug("%s is being processed without a reference.", txrm_name)
    return None

def _stitch_images(img_list, mosaic_xy_shape, fast_axis):
    slow_axis = 1 - fast_axis
    logging.debug("Fast axis: %i", fast_axis)
    
    num_images = len(img_list)
    expected_num_images = mosaic_xy_shape[0] * mosaic_xy_shape[1]
    if expected_num_images > num_images:  # Pad with zeros if mosaic interrupted
        img_list.extend([np.zeros((img_list[0].shape), dtype=img_list[0].dtype)] * (expected_num_images - num_images))
    
    fast_stacked_list = []
    for i in range(mosaic_xy_shape[slow_axis]):
        idx_start = mosaic_xy_shape[fast_axis] * i
        idx_end = mosaic_xy_shape[fast_axis] * (i + 1)
        logging.debug("Stacking images %i to %i along axis %i", idx_start, (idx_end - 1), fast_axis)
        fast_stacked_list.append(np.concatenate(img_list[idx_start: idx_end], axis=fast_axis))
    # Must be output in a list for iterative tiff saving
    final_image = np.concatenate(fast_stacked_list, axis=slow_axis)
    logging.info("Mosaic stitched into image with shape: %s", final_image.shape)
    return [final_image]


def manual_save(tiff_file, image, metadata=None):
    tiff_path = Path(tiff_file)

    if metadata is not None:
        metadata.image().set_Name(tiff_path.name)
        metadata = metadata.to_xml().encode()

    tiff_dir = tiff_path.resolve().parent
    tiff_dir.mkdir(parents=True, exist_ok=True)
    num_frames = len(image)
    logging.info("Saving image as %s with %i frames", tiff_path.name, num_frames)

    with tf.TiffWriter(str(tiff_path)) as tif:
        tif.save(image[0], photometric='minisblack', description=metadata, metadata={'axes':'XYZCT'})
        for i in range(1, num_frames):
            tif.save(image[i], photometric='minisblack', metadata={'axes':'XYZCT'})

def create_ome_metadata(ole, image_list, filename=None):
    # Get image dimensions
    # X, Y, number of frames (T in tilt series):
    dimensions = (*image_list[0].shape, len(image_list))
    str_dtype = str(image_list[0].dtype)

    # Get metadata variables from ole file:
    exposures = txrm_wrapper.extract_multiple_exposure_times(ole)

    pixel_size = txrm_wrapper.extract_pixel_size(ole) * 1.e3  # micron to nm

    physical_img_sizes = []
    physical_img_sizes.append(pixel_size * dimensions[0])
    physical_img_sizes.append(pixel_size * dimensions[1])

    x_positions = [coord * 1.e3 for coord in txrm_wrapper.extract_x_coords(ole)]  # micron to nm
    y_positions = [coord * 1.e3 for coord in txrm_wrapper.extract_y_coords(ole)]  # micron to nm

    date_time = datetime.datetime.now().isoformat()  # formatted as: "yyyy-mm-ddThh:mm:ss"
    ox = OMEXML()

    image = ox.image()
    image.set_ID("0")
    image.set_AcquisitionDate(date_time)
    if filename:
        image.set_Name(filename)

    pixels = image.Pixels
    pixels.set_DimensionOrder("XYTZC")
    pixels.set_ID("0")
    pixels.set_PixelType(str_dtype)
    pixels.set_SizeX(dimensions[0])
    pixels.set_SizeY(dimensions[1])
    pixels.set_SizeT(dimensions[2])
    pixels.set_SizeZ(1)
    pixels.set_SizeC(1)
    pixels.set_PhysicalSizeX(physical_img_sizes[0])
    pixels.set_PhysicalSizeXUnit("nm")
    pixels.set_PhysicalSizeY(physical_img_sizes[1])
    pixels.set_PhysicalSizeYUnit("nm")

    pixels.set_plane_count(dimensions[2])
    pixels.set_tiffdata_count(dimensions[2])

    channel = pixels.Channel(0)
    channel.set_ID("Channel:0:0")
    channel.set_Name("C:0")

    if (ole.exists("ImageInfo/MosiacRows") and ole.exists("ImageInfo/MosiacColumns")):
        mosaic_rows = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacRows")
        mosaic_columns = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacColumns")
        # Checks whether the output is a single frame and a mosaic
        # (mosaics should be stitched, and therefore one layer, by now):
        if (mosaic_rows > 0 and mosaic_columns > 0 and dimensions[2] == 1):
            # Calculates:
            # - Mean exposure, throwing away any invalid 0 values
            # - The centre of the stitched mosaic image (as opposed to the centre of a single tile)
            # Both should be returned as a list to reduce changes to the next section.
            exposures = [np.mean([exp for exp in exposures if exp != 0])]
            # The mosaic centre is found by taking the first x & y positions (centre of the first tile,
            # which is the bottom-left in the mosaic), taking away the distance between this and the
            # bottom-left corner, then adding the distance to the centre of the mosaic (calculated using pixel size).
            #
            # More verbosely:
            # The physical size of the stitched mosaic is divided by the rows/columns (columns for x, rows for y).
            # This finds the physical size of a single tile. This is then halved, finding in the physical
            # distance (x, y) from between a corner and the centre of a tile. Then this distance is taken from the
            # (x, y) stage coordinates of the first tile to get the stage coordinates of the bottom-left of the mosaic.
            # Half the physical size of the stitched mosaic is added to this, resulting in the in stage coordinates
            # of the mosaic centre.
            x_positions = [x_positions[0] + (1. - 1. / mosaic_columns) * (physical_img_sizes[0] / 2.)]
            y_positions = [y_positions[0] + (1. - 1. / mosaic_rows) * (physical_img_sizes[1] / 2.)]
            # # NOTE: the number of mosaic rows & columns and the pixel size are all written before acquisition but
            # the xy positions are written during, so only the first frame can be relied upon to have an xy
            # position.

    # Run checks to make sure the value lists are long enough
    exp_len_diff = dimensions[2] - len(exposures)
    if exp_len_diff > 0:
        logging.error("Not enough exposure values for each plane (%i vs %i). Adding zeros to the later planes.",
                        len(exposures), dimensions[2])
        for _ in range(exp_len_diff):
            exposures.append(0)

    x_len_diff = dimensions[2] - len(x_positions)
    if x_len_diff > 0:
        logging.error("Not enough x values for each plane (%i vs %i). Adding zeros to the later planes.",
                        len(x_positions), dimensions[2])
        for _ in range(x_len_diff):
            x_positions.append(0)

    y_len_diff = dimensions[2] - len(y_positions)
    if y_len_diff > 0:
        logging.error("Not enough y values for each plane (%i vs %i). Adding zeros to the later planes.",
                        len(y_positions), dimensions[2])
        for _ in range(y_len_diff):
            y_positions.append(0)

    # Add plane/tiffdata for each plane in the stack
    for count in range(dimensions[2]):
        plane = pixels.Plane(count)
        plane.set_ExposureTime(exposures[count])
        plane.set_TheZ(count)
        plane.set_TheC(0)
        plane.set_TheT(0)

        plane.set_PositionXUnit("nm")
        plane.set_PositionYUnit("nm")
        plane.set_PositionZUnit("reference frame")
        plane.set_PositionX(x_positions[count])
        plane.set_PositionY(y_positions[count])
        plane.set_PositionZ(count)

        tiffdata = pixels.tiffdata(count)
        tiffdata.set_FirstC(0)
        tiffdata.set_FirstZ(count)
        tiffdata.set_FirstT(0)
        tiffdata.set_IFD(count)
        tiffdata.set_plane_count(1)

    return ox

class TxrmToImage:

    def __init__(self, dtype=DEFAULT_NP_DTYPE):
        """
        Arguments:
            dtype = np.uint16 - accepts datatype strings or numpy types (e.g. np.uint16)
        """
        if issubclass(dtype, np.number):
            self.str_dtype = str(np.dtype(np.uint16))
            self.np_dtype = dtype
        elif isinstance(dtype, str):
            self.str_dtype = dtype
            self.np_dtype = getattr(np, dtype)
        else:
            logging.error("Invalid dtype '%s'. Reverting to default '%s'", dtype, DEFAULT_STR_DTYPE)
            self.str_dtype = DEFAULT_STR_DTYPE
            self.np_dtype = DEFAULT_NP_DTYPE
        self.image_output = None
        self.ome_metadata = None

    def convert(self, txrm_file, custom_reference=None, ignore_reference=False):
        with OleFileIO(str(txrm_file)) as ole:
            images = txrm_wrapper.extract_all_images(ole)
            reference = _get_reference(ole, txrm_file.name, custom_reference, ignore_reference)
            if reference is not None:
                self.image_output = _apply_reference(images, reference, self.np_dtype)
            else:
                self.image_output = [image for image in np.around(images).astype(self.np_dtype)]
            if (len(self.image_output) > 1
                    and ole.exists("ImageInfo/MosiacRows")
                    and ole.exists("ImageInfo/MosiacColumns")):
                    # Version 13 style mosaic:
                mosaic_rows = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacRows")
                mosaic_cols = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacColumns")
                if mosaic_rows != 0 and mosaic_cols != 0:
                    self.image_output = _stitch_images(self.image_output, (mosaic_cols, mosaic_rows), 1)
            # Create metadata
            self.ome_metadata = create_ome_metadata(ole, self.image_output)

    def get_image_and_metadata(self):
        if self.image_output is None:
            logging.warning("Image has not been converted, returning (None, None)")
        return self.image_output, self.ome_metadata


    def save(self, tiff_file):
        if (self.image_output is not None) and (self.ome_metadata is not None):
            manual_save(tiff_file, self.image_output, self.ome_metadata)
        else:
            logging.error("Nothing to save! Please convert the image first")
            raise IOError("Nothing to save! Please convert the image first")



import logging
import sys

level_dict = {
    "1": logging.DEBUG,
    "2": logging.INFO,
    "3": logging.WARNING,
    "4": logging.ERROR,
    "5": logging.CRITICAL,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    }


def create_logger(input_level):
    try:
        level = level_dict[input_level]
    except Exception as e:
        logging.error(e)

    logging_format = logging.Formatter("%(levelname)-8s %(message)s")
    handler = logging.StreamHandler(sys.stdout)

    logger = logging.getLogger(__name__)

    logger.setLevel(level)
    handler.setFormatter(logging_format)
    logger.addHandler(handler)



import unittest
from unittest.mock import MagicMock, patch
from numpy.testing import assert_array_equal
from datetime import datetime

from parameterized import parameterized

import numpy as np
import os
import logging
import sys
from pathlib import Path
from functools import partial
from shutil import rmtree
from time import time, sleep
from olefile import OleFileIO

from txrm2tiff.txrm_to_image import TxrmToImage, _get_reference, _apply_reference, create_ome_metadata


class TestTxrmToImageSimple(unittest.TestCase):

    def test_divides_images_by_reference(self):
        num_images = 5
        images = [np.array([[0, 2, 4], [6, 8, 10]])] * num_images
        reference = np.arange(6).reshape(2, 3)
        resultant_images = _apply_reference(images, reference, np.uint16)
        self.assertEqual(len(resultant_images), num_images, msg="The result is the wrong length")
        expected_image = np.array([[0, 200, 200], [200, 200, 200]])
        for image in resultant_images:
            assert_array_equal(image, expected_image, err_msg="The result does not match the expected result")
 
    @patch('txrm2tiff.txrm_to_image.txrm_wrapper')
    def test_metadata_created_correctly(self, mocked_extractor):
        dims = (45, 40, 1)
        dtype = "uint16"
        test_image = [np.zeros(dims[:2], dtype=dtype)] * dims[2]
        mosaic_rows = 2
        mosaic_cols = 3

        ole = MagicMock()
        ole.exists.return_value = True
        mocked_extractor.extract_multiple_exposure_times.return_value = [2., 3., 4., 5., 0., 0.]
        mocked_extractor.extract_pixel_size.return_value = 0.005
        mocked_extractor.extract_x_coords.return_value = [7.5, 22.5, 37.5] * 2
        mocked_extractor.extract_y_coords.return_value = [10, 10, 10, 20, 20, 20]
        mocked_extractor.read_imageinfo_as_int.side_effect = [mosaic_rows, mosaic_cols]

        ome_metadata = create_ome_metadata(ole, test_image)
        self.assertEqual(ome_metadata.image().Pixels.get_SizeX(), dims[0])
        self.assertEqual(ome_metadata.image().Pixels.get_SizeY(), dims[1])
        self.assertEqual(ome_metadata.image().Pixels.get_SizeT(), dims[2])
        self.assertEqual(ome_metadata.image().Pixels.get_PixelType(), dtype)

    @patch('txrm2tiff.txrm_to_image.txrm_wrapper')
    def test_mosaic_exposure_averaged(self, mocked_extractor):
        dims = (45, 40, 1)
        dtype = "uint16"
        test_image = [np.zeros(dims[:2], dtype=dtype)] * dims[2]
        mosaic_rows = 2
        mosaic_cols = 3
        exposure_times = [2., 3., 4., 5., 0., 0.]
        # 0 exposures should be ignored, if they exist, as these will be interrupted frames
        expected_exposure = 3.5

        ole = MagicMock()
        ole.exists.return_value = True
        mocked_extractor.extract_multiple_exposure_times.return_value = exposure_times
        mocked_extractor.extract_pixel_size.return_value = 0.005
        mocked_extractor.extract_x_coords.return_value = [7.5, 22.5, 37.5] * 2
        mocked_extractor.extract_y_coords.return_value = [10, 10, 10, 20, 20, 20]
        mocked_extractor.read_imageinfo_as_int.side_effect = [mosaic_rows, mosaic_cols]

        ome_metadata = create_ome_metadata(ole, test_image)
        self.assertEqual(ome_metadata.image().Pixels.Plane(0).get_ExposureTime(), expected_exposure)

    @patch('txrm2tiff.txrm_to_image.txrm_wrapper')
    def test_mosaic_centre_found_correctly(self, mocked_extractor):
        dims = (45, 40, 1)
        dtype = "uint16"
        test_image = [np.zeros(dims[:2], dtype=dtype)] * dims[2]
        pixel_size = 0.005
        mosaic_cols = 3
        mosaic_rows = 2
        exposure_times = [0., 0., 2., 3., 4., 5.]
        offset = [3., -2.]
        # Multiplier of 1.e3 required as units from xrm files are micrometres and the output should be in nanometres
        expected_centre = ((22.5 + offset[0]) * pixel_size * 1.e3, (20. + offset[1]) * pixel_size * 1.e3)

        
        image_divider = MagicMock()
        txrm_converter = TxrmToImage()
        ole = MagicMock()
        ole.exists.return_value = True
        mocked_extractor.extract_multiple_exposure_times.return_value = exposure_times
        mocked_extractor.extract_pixel_size.return_value = pixel_size

        # This should only need the coords of the first frame as mosaic may not complete
        mocked_extractor.extract_x_coords.return_value = [
            (7.5 + offset[0]) * pixel_size, 0, 0, 0, 0, 0]
        mocked_extractor.extract_y_coords.return_value = [
            (10 + offset[1]) * pixel_size, 0, 0, 0, 0, 0]

        mocked_extractor.read_imageinfo_as_int.side_effect = [mosaic_rows, mosaic_cols]

        ome_metadata = create_ome_metadata(ole, test_image)
        plane = ome_metadata.image().Pixels.Plane(0)
        ome_centre = (float(plane.get_PositionX()), float(plane.get_PositionY()))
        [self.assertAlmostEqual(ome, expected) for ome, expected in zip(ome_centre, expected_centre)]

    def test_save_before_convert(self):
        with self.assertRaises(IOError):
            TxrmToImage().save("output_path")

    @patch('txrm2tiff.txrm_to_image.isOleFile', MagicMock(return_value=True))
    @patch('txrm2tiff.txrm_to_image.OleFileIO')
    @patch('txrm2tiff.txrm_wrapper.extract_all_images')
    def test_get_reference_custom_median(self, mocked_extractor, mocked_olefile):
        custom_reference = []
        ole = MagicMock()
        for i in range(1, 5):
            custom_reference.append(np.full((5,5), i))
        custom_reference.append(np.full((5,5), 7))
        mocked_extractor.return_value = custom_reference
        ref_ole = MagicMock()
        mocked_olefile.return_value = ref_ole
        ref = _get_reference(ole, "txrm_name", ref_ole, ignore_reference=False)
        assert_array_equal(ref, custom_reference[2])

    @patch('txrm2tiff.txrm_to_image.isOleFile', MagicMock(return_value=True))
    @patch('txrm2tiff.txrm_to_image.OleFileIO')
    @patch('txrm2tiff.txrm_wrapper.extract_all_images') 
    def test_get_reference_custom_flat(self, mocked_extractor, mocked_olefile):
        ole = MagicMock()
        custom_reference = [np.full((5,5), 4)]
        mocked_extractor.return_value = custom_reference
        ref_ole = MagicMock()
        mocked_olefile.return_value = ref_ole
        ref = _get_reference(ole, "txrm_name", ref_ole, ignore_reference=False)
        assert_array_equal(ref, custom_reference[0])


base_path = Path("/dls/science/groups/das/ExampleData/B24_test_data/data/2019/cm98765-1")
raw_path = base_path / "raw"
xm10_path = raw_path / "XMv10"
xm13_path = raw_path / "XMv13"

test_files = [
    (xm13_path / 'Xray_mosaic_v13.xrm', ),
    (xm13_path / 'Xray_mosaic_v13_interrupt.xrm', ),
    (xm13_path / 'Xray_single_v13.xrm', ),
    (xm13_path / 'tomo_v13_full.txrm', ),
    (xm13_path / 'tomo_v13_full_noref.txrm', ),
    (xm13_path / 'tomo_v13_interrupt.txrm', ),
    (xm13_path / 'VLM_mosaic_v13.xrm', ),
    (xm13_path / 'VLM_mosaic_v13_interrupt.xrm', ),
    (xm10_path / '12_Tomo_F4D_Area1_noref.txrm', ),
    (xm10_path / 'VLM_mosaic.xrm', ),
    (xm10_path / 'test_tomo2_e3C_full.txrm', ),
    (xm10_path / 'Xray_mosaic_F5A.xrm', ),
]


@unittest.skipIf(not base_path.exists(), "dls paths cannot be accessed")
class TestTxrmToImageWithFiles(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.base_path = base_path
        cls.raw_path = raw_path
        cls.processed_path = cls.base_path / "processed"
        if cls.processed_path.exists():
            cleanup_timeout = 40
            # rmtree is asynchronous, so a wait may be required:
            rmtree(cls.processed_path, ignore_errors=True)
            start = time()
            while os.path.exists(cls.processed_path) and (time() - start) < cleanup_timeout:
                sleep(1)

    def setUp(self):
        self.processed_path.mkdir()
        self.assertTrue(self.processed_path.exists(), msg="Processed folder not correctly created")
    
    def tearDown(self):
        cleanup_timeout = 40
        # rmtree is asynchronous, so a wait may be required:
        rmtree(self.processed_path, ignore_errors=True)
        start = time()
        while os.path.exists(self.processed_path) and (time() - start) < cleanup_timeout:
            sleep(1)

        self.assertFalse(self.processed_path.exists(), msg="Processed folder not correctly removed")

    @parameterized.expand(test_files)
    def test_converts_to_tiff(self, test_file):
        logging.debug("Running with file %s", test_file)
        output_path = self.processed_path / test_file.relative_to(self.raw_path).with_suffix('.ome.tiff')
        converter = TxrmToImage()
        # Make processed/ subfolders:
        output_path.parent.mkdir(parents=True, exist_ok=True)

        converter.convert(test_file, None, False)
        converter.save(output_path)

        self.assertTrue(output_path.exists())


suite1 = unittest.TestLoader().loadTestsFromTestCase(TestTxrmToImageSimple)
suite2 = unittest.TestLoader().loadTestsFromTestCase(TestTxrmToImageWithFiles)
alltests = unittest.TestSuite([suite1, suite2])

if __name__ == "__main__":
    unittest.TextTestRunner().run(alltests)